window.Gleam.initSearch([{"doc":"lamb","title":"lamb","content":"# lamb\n\n[![Package Version](https://img.shields.io/hexpm/v/lamb)](https://hex.pm/packages/lamb)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/lamb/)\n\nA gleam library for operating and querying ETS tables.\n\n```gleam\nimport gleam/list\nimport lamb.{Set, Private}\nimport lamb/query.{v, i, a} as q\n\npub fn main() {\n  // Create a table and insert 5 records.\n  let assert Ok(table) = lamb.create_table(Set, Private, &quot;test_table&quot;)\n\n  lamb.insert(table, [\n    #(&quot;a&quot;,  1),\n    #(&quot;b&quot;,  2),\n    #(&quot;c&quot;,  3),\n    #(&quot;d&quot;,  4),\n    #(&quot;e&quot;,  5),\n  ])\n\n  // Retrieve just 1 record.\n  let assert Ok(3) = lamb.get(table, &quot;c&quot;)\n\n  // This query syntax builds a matchspec to make queries on the table.\n  let query =\n    q.new()\n    |&gt; q.bind(#(v(0), v(1)))\n\n  // Retrieve all rows but only return the record\n  let _records = lamb.all(table, query |&gt; q.map(v(1)))\n\n  // Retrieve all rows but only return the index\n  let _records = lamb.all(table, query |&gt; q.map(v(0)))\n\n  // Retrieve all records in batches of 2.\n  let assert Records([_, _] as a, step) = lamb.partial(table, by: 2, where: q.new())\n  let assert Records([_, _] as b, step) = lamb.continue(step)\n  let assert End([_] as c) = lamb.continue(step)\n}\n```\n\nFurther documentation can be found at &lt;https://hexdocs.pm/lamb&gt;.\n\n## Development\n\n```sh\ngleam run   # Run the project\ngleam test  # Run the tests\n```\n\n---\n\nWork in progress notes.\n\n# Table API\n\nCurrently viewing the table creating in terms of protection levels:\n  - `Private`, private table, not named.\n  - `Protected`, protected table, named.\n  - `Public`, public table, named.\n\nIf we would like more precision than this, maybe having an `options` helper with default options would\nhelp to tinker with tables:\n\n```gleam\npub fn options() -&gt; Options {\n  https://www.erlang.org/doc/apps/stdlib/ets.html#info/1\n  todo\n}\n```\n\nStill need to figure out what is going to be the API for differentiating between `set` and `bag` tables.\n\n# Query API\n\nMatchspecs are composed by a Tuple of arity 3 called a `MatchFunction`:\n\n- A `Head` that contains the shape of the data we want to match to, as well as variable declarations.\n- A list of `Condition` expressions that can help filter data through predicates.\n- A `Body` that declares the shape and variables we&#39;d like to output from the `MatchFunction`.\n\n```erlang\n{Head, [Condition], [Body]}\n```\n\nOperating on these 3 pieces may be tackled in different ways.\n\n## Alternative 1\n\nHave an API that composes matchspecs together with a builder pattern:\n\n```gleam\nlet query =\n  q.new()                                  // Builds a basic default match function.\n  |&gt; q.bind(Person)                        // Modifies the head to match the constructor.\n  |&gt; q.match(field: 5, &quot;Citizen&quot;)          // Modifies the head to match the value.\n  |&gt; q.where(field: 3, op: &quot;&gt;=&quot;, than: 18) // Adds a condition expression.\n\nquery |&gt; q.select(object())                // Returns the whole object #(index, record).\nquery |&gt; q.select(index())                 // Returns just the index of record.\nquery |&gt; q.select(#(v(2), v(1)))           // Returns the 2nd and 1st variables in a tuple.\nquery |&gt; q.select(False)                   // Returns false for each matching record.\nquery |&gt; q.select(Driver)                  // Maps the variables to a constructor.\nquery |&gt; q.select(fn(last, first) {        // Alternative mapper to above.\n  Driver(last, first)\n})\n```\n\n## Alternative 2\n\nBuild a &quot;nice&quot; parser to compose complex queries:\n\n```gleam\n\nlet query = &quot;\n  from Person\n  select 2, 1\n  where &#39;Citizen&#39; = 5\n  where 3 &gt;= 18\n  select #(0, 1, 2)\n&quot;\n```\n\n## Alternative 3\n\nJust build a more straightforward mapper for matchspecs.\n\n```gleam\nlet head = #(any(), record(&quot;Person&quot;, v(1), v(2), v(3), &quot;Citizen&quot;))\nlet condition = #(&quot;&gt;=&quot;, v(3), 18)\nlet body = body(record(&quot;Driver&quot;, v(2), v(1)))\nlet query_driver_elegibility = #(head, [condition], [body])\nlamb.search(table, [query_driver_elegibility])\n```\n\n## Notes on types and validation\n\nNone of these alternatives currently provide a good way of enforcing types but are meant to fail gracefully if\nthere are errors. Querying so far is a &quot;dynamic&quot; operation.\n\nAn alternative would be to run validations at runtime with the help of a schema but that would\nbe quite a big lift for my current purposes.\n\nMaybe another way is to provide a validator at init time, but this would exclusively check the validity of the\nmatchspecs.\n\n# Checkout these ETS APIs\n\n* https://www.erlang.org/doc/apps/stdlib/ets.html#select_delete/2\n* https://www.erlang.org/doc/apps/stdlib/ets.html#fun2ms/1\n* https://www.erlang.org/doc/apps/stdlib/ets.html#test_ms/2\n* https://www.erlang.org/doc/apps/stdlib/ets.html#table/2\n* https://www.erlang.org/doc/apps/stdlib/qlc\n* https://www.erlang.org/doc/apps/stdlib/erl_parse.html#abstract/1\n","url":"index.html"},{"doc":"lamb/query","title":"Query","content":"pub type Query(index, record)\n\n\nSynonyms:\nlamb/query.Query\nlamb/query Query","url":"lamb/query.html#Query"},{"doc":"lamb/query","title":"a","content":"pub fn a(name: String) -&gt; Atom\n\nSynonyms:\nlamb/query.a\nlamb/query a","url":"lamb/query.html#a"},{"doc":"lamb/query","title":"against","content":"pub fn against(query: Query(a, b), row: c) -&gt; Result(d, Nil)\n\nSynonyms:\nlamb/query.against\nlamb/query against","url":"lamb/query.html#against"},{"doc":"lamb/query","title":"bind","content":"pub fn bind(query: Query(a, b), with shape: c) -&gt; Query(a, b)\n\nSynonyms:\nlamb/query.bind\nlamb/query bind","url":"lamb/query.html#bind"},{"doc":"lamb/query","title":"filter","content":"pub fn filter(\n  query: Query(a, b),\n  with condition: c,\n) -&gt; Query(a, b)\n\nSynonyms:\nlamb/query.filter\nlamb/query filter","url":"lamb/query.html#filter"},{"doc":"lamb/query","title":"i","content":"pub fn i() -&gt; Atom\n\nSynonyms:\nlamb/query.i\nlamb/query i","url":"lamb/query.html#i"},{"doc":"lamb/query","title":"map","content":"pub fn map(query: Query(a, b), with shape: c) -&gt; Query(a, b)\n\nSynonyms:\nlamb/query.map\nlamb/query map","url":"lamb/query.html#map"},{"doc":"lamb/query","title":"new","content":"pub fn new() -&gt; Query(a, b)\n\nSynonyms:\nlamb/query.new\nlamb/query new","url":"lamb/query.html#new"},{"doc":"lamb/query","title":"r","content":"pub fn r(name: String, tuple: a) -&gt; b\n\nSynonyms:\nlamb/query.r\nlamb/query r","url":"lamb/query.html#r"},{"doc":"lamb/query","title":"t1","content":"pub fn t1(a: a) -&gt; #(#(a))\n\nSynonyms:\nlamb/query.t1\nlamb/query t1","url":"lamb/query.html#t1"},{"doc":"lamb/query","title":"t2","content":"pub fn t2(a: a, b: b) -&gt; #(#(a, b))\n\nSynonyms:\nlamb/query.t2\nlamb/query t2","url":"lamb/query.html#t2"},{"doc":"lamb/query","title":"t3","content":"pub fn t3(a: a, b: b, c: c) -&gt; #(#(a, b, c))\n\nSynonyms:\nlamb/query.t3\nlamb/query t3","url":"lamb/query.html#t3"},{"doc":"lamb/query","title":"t4","content":"pub fn t4(a: a, b: b, c: c, d: d) -&gt; #(#(a, b, c, d))\n\nSynonyms:\nlamb/query.t4\nlamb/query t4","url":"lamb/query.html#t4"},{"doc":"lamb/query","title":"t5","content":"pub fn t5(a: a, b: b, c: c, d: d, e: e) -&gt; #(#(a, b, c, d, e))\n\nSynonyms:\nlamb/query.t5\nlamb/query t5","url":"lamb/query.html#t5"},{"doc":"lamb/query","title":"v","content":"pub fn v(at position: Int) -&gt; Atom\n\nSynonyms:\nlamb/query.v\nlamb/query v","url":"lamb/query.html#v"},{"doc":"lamb/query","title":"validate","content":"pub fn validate(\n  query: Query(a, b),\n) -&gt; Result(Query(a, b), List(String))\n\nSynonyms:\nlamb/query.validate\nlamb/query validate","url":"lamb/query.html#validate"},{"doc":"lamb/query","title":"lamb/query","content":"","url":"lamb/query.html"},{"doc":"lamb","title":"Access","content":"pub type Access {\n  Public\n  Protected\n  Private\n}\n\nPublic\n\n\nProtected\n\n\nPrivate\n\n\nSynonyms:\nlamb.Access\nlamb Access","url":"lamb.html#Access"},{"doc":"lamb","title":"Partial","content":"pub type Partial(record) {\n  Records(List(record), Step)\n  End(List(record))\n}\n\nRecords(List(record), Step)\n\n\nEnd(List(record))\n\n\nSynonyms:\nlamb.Partial\nlamb Partial","url":"lamb.html#Partial"},{"doc":"lamb","title":"Step","content":"pub type Step\n\n\nSynonyms:\nlamb.Step\nlamb Step","url":"lamb.html#Step"},{"doc":"lamb","title":"Store","content":"pub type Store {\n  Set\n}\n\nSet\n\n\nSynonyms:\nlamb.Store\nlamb Store","url":"lamb.html#Store"},{"doc":"lamb","title":"Table","content":"pub opaque type Table(index, record)\n\n\nSynonyms:\nlamb.Table\nlamb Table","url":"lamb.html#Table"},{"doc":"lamb","title":"all","content":"pub fn all(\n  from table: Table(a, b),\n  where query: Query(a, b),\n) -&gt; List(c)\n\nSynonyms:\nlamb.all\nlamb all","url":"lamb.html#all"},{"doc":"lamb","title":"batch","content":"pub fn batch(\n  from table: Table(a, b),\n  by limit: Int,\n  where query: Query(a, b),\n) -&gt; Partial(c)\n\nSynonyms:\nlamb.batch\nlamb batch","url":"lamb.html#batch"},{"doc":"lamb","title":"continue","content":"pub fn continue(step: Step) -&gt; Partial(a)\n\nSynonyms:\nlamb.continue\nlamb continue","url":"lamb.html#continue"},{"doc":"lamb","title":"count","content":"pub fn count(\n  from table: Table(a, b),\n  where query: Query(a, b),\n) -&gt; Int\n\nSynonyms:\nlamb.count\nlamb count","url":"lamb.html#count"},{"doc":"lamb","title":"create_table","content":"pub fn create_table(\n  store: Store,\n  access: Access,\n  name: String,\n) -&gt; Result(Table(a, b), Nil)\n\nSynonyms:\nlamb.create_table\nlamb create_table","url":"lamb.html#create_table"},{"doc":"lamb","title":"delete","content":"pub fn delete(\n  table: Table(a, b),\n  where query: Query(a, b),\n) -&gt; Int\n\nSynonyms:\nlamb.delete\nlamb delete","url":"lamb.html#delete"},{"doc":"lamb","title":"delete_table","content":"pub fn delete_table(table: Table(a, b)) -&gt; Nil\n\nSynonyms:\nlamb.delete_table\nlamb delete_table","url":"lamb.html#delete_table"},{"doc":"lamb","title":"from_name","content":"pub fn from_name(name: String) -&gt; Result(Table(a, b), Nil)\n\nSynonyms:\nlamb.from_name\nlamb from_name","url":"lamb.html#from_name"},{"doc":"lamb","title":"get","content":"pub fn get(table: Table(a, b), index: a) -&gt; Result(b, Nil)\n\nSynonyms:\nlamb.get\nlamb get","url":"lamb.html#get"},{"doc":"lamb","title":"insert","content":"pub fn insert(table: Table(a, b), rows: List(#(a, b))) -&gt; Nil\n\nSynonyms:\nlamb.insert\nlamb insert","url":"lamb.html#insert"},{"doc":"lamb","title":"is_alive","content":"pub fn is_alive(table: Table(a, b)) -&gt; Bool\n\nSynonyms:\nlamb.is_alive\nlamb is_alive","url":"lamb.html#is_alive"},{"doc":"lamb","title":"update","content":"pub fn update(\n  table: Table(a, b),\n  where query: Query(a, b),\n) -&gt; Int\n\nSynonyms:\nlamb.update\nlamb update","url":"lamb.html#update"},{"doc":"lamb","title":"lamb","content":"","url":"lamb.html"},{"doc":"lamb/experimental","title":"lamb/experimental","content":"","url":"lamb/experimental.html"}]);